=begin

Xlib for Ruby
Copyright (c) 2001 by Mathieu Bouchard
Under Ruby Dual License (see COPYING* files)

Derived from:
	XConsortium: Xlib.h /main/119 1996/09/28 16:35:29 rws 
	XFree86: xc/lib/X11/Xlib.h,v 3.10.2.5 1999/07/26 06:54:33 hohndel Exp
	(Under MIT license)

Report all translation errors to Mathieu Bouchard <matju@cam.org>

=end

=begin

Xlib is the standard bloated library that operates the client side of an
X Window System protocol connection.

I translated the original C declarations into Ruby data structures. Then
I wrote a compiler from those data structures to C.

=end

# Structures and symbols starting with "_" are private to the library.

#include <X11/X.h>
#/* applications should not depend on these two headers being included! */
#include <X11/Xfuncproto.h>
#include <X11/Xosdefs.h>

# typedef char *XPointer;

ExtData = C_Struct(
	'Extensions need a way to hang private data on some structures.',
[
	[:number,'int','number returned by XRegisterExtension'],
	[:next,'XExtData*','next item on list of data for structure'],
	[:free_private,'int(*)(Xpointer)','called to free private storage'],
	[:private_data,'XPointer','data private to this extension.'],
])

XExtCodes = C_Struct(
	'This file contains structures used by the extension mechanism.
	public to extension, cannot be changed
	',
[
	[:extension,'int','extension number'],
	[:major_opcode,'int','major op-code assigned by server'],
	[:first_event,'int','first event number for the extension'],
	[:first_error,'int','first error number for the extension'],
])

XPixmapFormatValues = C_Struct(
	'Data structure for retrieving info about pixmap formats.',
[
	[:depth,'int',''],
	[:bits_per_pixel,'int',''],
	[:scanline_pad,'int',''],
])

XGCValues = C_Struct(
	'Data structure for setting graphics context.',
[
	[:function,'int','logical operation'],
	[:plane_mask,'ulong','plane mask'],
	[:foreground,'ulong','foreground pixel'],
	[:background,'ulong','background pixel'],
	[:line,'int','line_width'],
	[:line_style,'int','LineSolid, LineOnOffDash, LineDoubleDash'],
	[:cap_style,'int','CapNotLast, CapButt, CapRound, CapProjecting'],
	[:join_style,'int','JoinMiter, JoinRound, JoinBevel'],
	[:fill_style,'int','FillSolid, FillTiled, FillStippled, FillOpaqueStippled'],
	[:fill_rule,'int','EvenOddRule, WindingRule'],
	[:arc_mode,'int','ArcChord, ArcPieSlice'],
	[:tile,'Pixmap','tile pixmap for tiling operations'],
	[:stipple,'Pixmap','stipple 1 plane pixmap for stipping'],
	[:ts_origin,'%Position','offset for tile or stipple operations'],
	[:font,'Font','default text font for text operations'],
	[:subwindow_mode,'int','ClipByChildren, IncludeInferiors'],
	[:graphics_exposures,'Bool','boolean, should exposures be generated'],
	[:clip_origin,'%Position','origin for clipping'],
	[:clip_mask,'Pixmap','bitmap clipping; other calls for rects'],
	[:dash_offset,'int','patterned/dashed line information'],
	[:dashes,'char',''],
])

_XGC = C_Struct(
	'Graphics context.  The contents of this structure are implementation
	dependent.  A GC should be treated as opaque by application code.',
[
	[:ext_data,'XExtData*','hook for extension to hang data'],
	[:gid,'GContext','protocol ID for graphics context'],
	# there is more to this structure, but it is private to Xlib
])

#!@#$ GC is a pointer to _XGC

Visual = C_Struct(
	'Visual structure; contains information about colormapping possible.',
[
	[:ext_data,'XExtData*','hook for extension to hang data'],
	[:visualid,'VisualID','visual id of this visual'],
	[:class,'int','class of screen (monochrome, etc.)'],
	[:mask,'%RGB_ulong','mask values'],
	[:bits_per_rgb,'int','log base 2 of distinct color values'],
	[:map_entries,'int','color map entries'],
])

Depth = C_Struct(
	'Depth structure; contains information for each possible depth.',
[
	[:depth,'int','this depth (Z) of the depth'],
	[:nvisuals,'int','number of Visual types at this depth'],
	[:visuals,'Visual*','list of visuals possible at this depth'],
])

Screen = C_Struct(
	'Information about the screen.  The contents of this structure are
	implementation dependent.  A Screen should be treated as opaque
	by application code.',
[
	[:ext_data,'XExtData *','hook for extension to hang data'],
	[:display,'_XDisplay *','back pointer to display structure'],
	[:root,'Window','Root window id.'],
	[:size,'%Size','width and height of screen'],
	[:msize,'%Size','width and height of  in millimeters'],
	[:ndepths,'int','number of depths possible'],
	[:depths,'Depth *','list of allowable depths on the screen'],
	[:root_depth,'int','bits per pixel'],
	[:root_visual,'Visual *','root visual'],
	[:default_gc,'GC','GC for the root root visual'],
	[:cmap,'Colormap','default color map'],
	[:white_pixel,'ulong',''],
	[:black_pixel,'ulong','White and Black pixel values'],
	[:max_maps,'int','max color maps'],
	[:min_maps,'int','min color maps'],
	[:backing_store,'int','Never, WhenMapped, Always'],
	[:save_unders,'Bool',''],
	[:root_input_mask,'long','initial root input mask'],
])

ScreenFormat = C_Struct(
	'Format structure; describes ZFormat data the screen will understand.',
[
	[:ext_data,'XExtData *','hook for extension to hang data'],
	[:depth,'','depth of this image format'],
	[:bits_per_pixel,'','bits/pixel at this depth'],
	[:scanline_pad,'','scanline must padded to this multiple'],
])

XSetWindowAttributes = C_Struct(
	'Data structure for setting window attributes.',
[
	[:background_pixmap,'Pixmap','background or None or ParentRelative'],
	[:background_pixel,'ulong','background pixel'],
	[:border_pixmap,'Pixmap','border of the window'],
	[:border_pixel,'ulong','border pixel value'],
	[:bit_gravity,'int','one of bit gravity values'],
	[:win_gravity,'int','one of the window gravity values'],
	[:backing_store,'int','NotUseful, WhenMapped, Always'],
	[:backing_planes,'ulong','planes to be preseved if possible'],
	[:backing_pixel,'ulong','value to use in restoring planes'],
	[:save_under,'Bool','should bits under be saved? (popups)'],
	[:event_mask,'long','set of events that should be saved'],
	[:do_not_propagate_mask,'long','set of events that should not propagate'],
	[:override_redirect,'Bool','boolean value for override-redirect'],
	[:colormap,'Colormap','color map to be associated with window'],
	[:cursor,'Cursor','cursor to be displayed (or None)'],
])

XWindowAttributes = C_Struct(
	'',
[
	[:position,'%Position','location of window'],
	[:size,'%Size','width and height of window'],
	[:border_width,'','border width of window'],
	[:depth,'','depth of window'],
	[:visual,'Visual *','the associated visual structure'],
	[:root,'Window','root of screen containing window'],
	[:class,'','InputOutput, InputOnly'],
	[:bit_gravity,'','one of bit gravity values'],
	[:win_gravity,'','one of the window gravity values'],
	[:backing_store,'','NotUseful, WhenMapped, Always'],
	[:backing_planes,'ulong','planes to be preserved if possible'],
	[:backing_pixel,'ulong','value to be used when restoring planes'],
	[:save_under,'Bool','boolean, should bits under be saved?'],
	[:colormap,'Colormap','color map to be associated with window'],
	[:map_installed,'Bool','boolean, is color map currently installed'],
	[:map_state,'','IsUnmapped, IsUnviewable, IsViewable'],
	[:all_event_masks,'long','set of events all people have interest in'],
	[:your_event_mask,'long','my event mask'],
	[:do_not_propagate_mask,'long','set of events that should not propagate'],
	[:override_redirect,'Bool','boolean value for override-redirect'],
	[:screen,'Screen *','back pointer to correct screen'],
])

XHostAddress = C_Struct(
	'Data structure for host setting; getting routines.',
[
	[:family,'','for example FamilyInternet'],
	[:length,'','length of address, in bytes'],
	[:address,'char*','pointer to where to find the bytes'],
])

XImage = C_Struct(
	'Data structure for "image" data, used by image manipulation routines.',
[
	[:size,'%Size','size of image'],
	[:xoffset,'','number of pixels offset in X direction'],
	[:format,'','XYBitmap, XYPixmap, ZPixmap'],
	[:data,'char *','pointer to image data'],
	[:byte_order,'','data byte order, LSBFirst, MSBFirst'],
	[:bitmap_unit,'','quant. of scanline 8, 16, 32'],
	[:bitmap_bit_order,'','LSBFirst, MSBFirst'],
	[:bitmap_pad,'','8, 16, 32 either XY or ZPixmap'],
	[:depth,'','depth of image'],
	[:bytes_per_line,'','accelarator to next line'],
	[:bits_per_pixel,'','bits per pixel (ZPixmap)'],
	[:mask,'%RGB_ulong','bits in z arrangment'],
	[:obdata,'XPointer','hook for the object routines to hang on'],
#	struct funcs {		'image manipulation routines'],
#!@#$
#	_XImage *(*create_image)(
#		_XDisplay*,Visual*,
#		uint'depth'],
#		int'format'],
#		int'offset'],
#		char*'data'],
#		[:size,'%Size_uint',''],
#		int'bitmap_pad'],,
#		int'bytes_per_line'],);
#	int (*destroy_image)(_XImage*);
#	ulong (*get_pixel)  (_XImage*,int,int)
#	int (*put_pixel)	(_XImage*,int,int,ulong)
#	_XImage *(*sub_image)(_XImage*,int,int,uint,uint)
#	int (*add_pixel)	(_XImage*,long)
#	} f;
])

XWindowChanges = C_Struct(
	'Data structure for XReconfigureWindow',
[
	[:position,'%Position'],
	[:size,'%size'],
	[:border_width,''],
	[:sibling,'Window'],
	[:stack_mode,''],
])

XColor = C_Struct(
	'Data structure used by color operations',
[
	[:pixel,'ulong'],
	[:red,'ushort'],
	[:green,'ushort'],
	[:blue,'ushort'],
	[:flags,'char','do_red, do_green, do_blue'],
	[:pad,'char'],
])

XSegment = C_Struct(
	'Data structures for graphics operations.  On most machines, these are
	congruent with the wire protocol structures, so reformatting the data
	can be avoided on these architectures.',
[
	[:x1,'short',''],
	[:y1,'short',''],
	[:x2,'short',''],
	[:y2,'short',''],
])

XPoint = C_Struct(
	'',
[
	[:x,'short',''],
	[:y,'short',''],
])

XRectangle = C_Struct(
	'',
[
	[:x,'short',''],
	[:y,'short',''],
	[:size,'%Size_ushort'],
])
	
XArc = C_Struct(
	'',
[
	[:x,'short',''],
	[:y,'short',''],
	[:size,'%Size_ushort'],
	[:angle1,'short'],
	[:angle2,'short'],
])
	
XKeyboardControl = C_Struct(
	'Data structure for XChangeKeyboardControl',
[
	[:key_click_percent],
	[:bell_percent],
	[:bell_pitch],
	[:bell_duration],
	[:led],
	[:led_mode],
	[:key],
	[:auto_repeat_mode,'On, Off, Default'],
])

XKeyboardState = C_Struct(
	'Data structure for XGetKeyboardControl',
[
	[:key_click_percent,''],
	[:bell_percent,''],
	[:bell_pitch,'uint'],
	[:bell_duration,''],
	[:led_mask,'ulong'],
	[:global_auto_repeat,''],
	[:auto_repeats,'char[32]'],
])

XTimeCoord = C_Struct(
	'Data structure for XGetMotionEvents. ',
[
	[:time,'Time'],
	[:x,'short'],
	[:y,'short'],

])

XModifierKeymap = C_Struct(
	'Data structure for X{Set,Get}ModifierMapping',
[
 	[:max_keypermod,'',"The server's max # of keys per modifier"],
 	[:modifiermap,'KeyCode*','An 8 by max_keypermod array of modifiers'],
])

Display = C_Struct(
	'Display datatype maintaining display specific data.
	The contents of this structure are implementation dependent.
	A Display should be treated as opaque by application code.',
[
	[:ext_data,'XExtData*','hook for extension to hang data'],
	[:private1,'_XPrivate *'],
	[:fd,'','Network socket'],
	[:private2,''],
	[:proto_major_version,'',"major version of server's X protocol"],
	[:proto_minor_version,'',"minor version of server's X protocol"],
	[:vendor,'char*','vendor of the server hardware'],
	[:private3,'XID'],
	[:private4,'XID'],
	[:private5,'XID'],
	[:private6,''],
	[:resource_alloc,'XID (*)(_XDisplay*)','allocator function'],
	[:byte_order,'','screen byte order, LSBFirst, MSBFirst'],
	[:bitmap_unit,'','padding and data requirements'],
	[:bitmap_pad,'','padding requirements on bitmaps'],
	[:bitmap_bit_order,'','LeastSignificant or MostSignificant'],
	[:nformats,'','number of pixmap formats in list'],
	[:pixmap_format,'ScreenFormat*','pixmap format list'],
	[:private8,'',int],
	[:release,'','release of the server'],
	[:private9,'_XPrivate*'],
	[:private10,'_XPrivate*'],
	[:qlen,'','Length of input event queue'],
	[:last_request_read,'ulong','seq number of last event read'],
	[:request,'ulong','sequence number of last request.'],
	[:private11,'XPointer'],
	[:private12,'XPointer'],
	[:private13,'XPointer'],
	[:private14,'XPointer'],
	[:max_request_size,'uint','maximum number 32 bit words in request'],
	[:db,'_XrmHashBucketRec*'],
	[:private15,'int (*)(_XDisplay*)'],
	[:display_name,'char *','"host:display" string used on this connect'],
	[:default_screen,'','default screen for operations'],
	[:nscreens,'','number of screens on this server'],
	[:screens,'Screen *','pointer to list of screens'],
	[:motion_buffer,'ulong','size of motion buffer'],
	[:private16,'ulong'],
	[:min_keycode,'','minimum defined keycode'],
	[:max_keycode,'','maximum defined keycode'],
	[:private17,'XPointer'],
	[:private18,'XPointer'],
	[:private19,''],
	[:xdefaults,'char *','contents of defaults from server'],
])

SameScreenFlag = [:same_screen,'Bool','same screen flag']

# 'there is more to this structure, but it is private to Xlib'

# Definitions of specific events.

# most any event starts like this.
XAnyEvent = C_Struct([
	[:type,'','event type number'],
	[:serial,'ulong','of last request processed by server'],
	[:send_event,'Bool','true if this came from a SendEvent request'],
	DPY,
	[:window,'Window','window on which event was requested in event mask'],
])

XKeyEvent = XAnyEvent.subtype([
	[:root,'Window','root window that the event occurred on'],
	[:subwindow,'Window','child window'],
	[:time,'Time','milliseconds'],
	[:position,'%Position','pointer x, y coordinates in event window'],
	[:position_root,'%Position','coordinates relative to root'],
	[:state,'uint','key or button mask'],
	[:keycode,'uint','detail'],
	SameScreenFlag,
])

XKeyPressedEvent = XKeyEvent.subtype
XKeyReleasedEvent = XKeyEvent.subtype

XButtonEvent = XAnyEvent.subtype([
	[:root,'Window','root window that the event occurred on'],
	[:subwindow,'Window','child window'],
	[:time,'Time','milliseconds'],
	[:position,'%Position','pointer x, y coordinates in event window'],
	[:position_root,'%Position','coordinates relative to root'],
	[:state,'uint','key or button mask'],
	[:button,'uint','detail'],
	SameScreenFlag,
])

XButtonPressedEvent = XButtonEvent.subtype
XButtonReleasedEvent = XButtonEvent.subtype

XMotionEvent = XAnyEvent.subtype([
	[:root,'Window','root window that the event occurred on'],
	[:subwindow,'Window','child window'],
	[:time,'Time','milliseconds'],
	[:position,'%Position','pointer x, y coordinates in event window'],
	[:position_root,'%Position','coordinates relative to root'],
	[:state,'uint','key or button mask'],
	[:is_hint,'char','detail'],
	SameScreenFlag,
])
typedef XMotionEvent XPointerMovedEvent;

XCrossingEvent = XAnyEvent.subtype([
	[:root,'Window','root window that the event occurred on'],
	[:subwindow,'Window','child window'],
	[:time,'Time','milliseconds'],
	[:position,'%Position','pointer x, y coordinates in event window'],
	[:position_root,'%Position','coordinates relative to root'],
	[:mode,'','NotifyNormal, NotifyGrab, NotifyUngrab'],
	[:detail,'','NotifyAncestor, NotifyVirtual, NotifyInferior, 
		NotifyNonlinear,NotifyNonlinearVirtual'],
	SameScreenFlag,
	[:focus,'Bool','boolean focus'],
	[:state,'uint','key or button mask'],
])
XEnterWindowEvent = XCrossingEvent.subtype
XLeaveWindowEvent = XCrossingEvent.subtype

XFocusChangeEvent = XAnyEvent.subtype([
	[:mode,'','NotifyNormal, NotifyGrab, NotifyUngrab'],
	[:detail,'','NotifyAncestor, NotifyVirtual, NotifyInferior, 
		NotifyNonlinear,NotifyNonlinearVirtual, NotifyPointer,
		NotifyPointerRoot, NotifyDetailNone'],
])
XFocusInEvent  = XFocusChangeEvent.subtype
XFocusOutEvent = XFocusChangeEvent.subtype

XKeymapEvent = XAnyEvent.subtype(
	'generated on EnterWindow and FocusIn  when KeyMapState selected',
[
	[:key_vector,'char[32]'],
])

XExposeEvent = XAnyEvent.subtype([
	[:rectangle,'%Rectangle','position and size in window'],
	[:count,'','if non-zero, at least this many more'],
])

XGraphicsExposeEvent = XAnyEvent.subtype([
	# first window was a drawable instead
	[:rectangle,'%Rectangle','position and size in window'],
	[:count,'','if non-zero, at least this many more'],
	[:major_code,'','core is CopyArea or CopyPlane'],
	[:minor_code,'','not defined in the core'],
])

XNoExposeEvent = XAnyEvent.subtype([
	# first window was a drawable instead
	[:major_code,'','core is CopyArea or CopyPlane'],
	[:minor_code,'','not defined in the core'],
])

XVisibilityEvent = XAnyEvent.subtype([
	[:state,'','Visibility state'],
])

XCreateWindowEvent = XAnyEvent.subtype([
	# first window was called parent instead
	[:window,'Window','window id of window created'],
	[:rectangle,'%Rectangle','position and size of window'],
	[:border_width,'','border width'],
	[:override_redirect,'Bool','creation should be overridden'],
])

XDestroyWindowEvent = XAnyEvent.subtype([
	# first window was called event instead
	[:window,'Window'],
])

XUnmapEvent = XAnyEvent.subtype([
	# first window was called event instead
	[:window,'Window'],
	[:from_configure,'Bool'],
])

XMapEvent = XAnyEvent.subtype([
	# first window was called event instead
	[:window,'Window'],
	[:override_redirect,'Bool','boolean, is override set...'],
])

XMapRequestEvent = XAnyEvent.subtype([
	# first window was called parent instead
	[:window,'Window'],
])

XReparentEvent = XAnyEvent.subtype([
	# first window was called event instead
	[:window,'Window'],
	[:parent,'Window'],
	[:position,'%Position'],
	[:override_redirect,'Bool'],
])

XConfigureEvent = XAnyEvent.subtype([
	# first window was called event instead
	[:window,'Window'],
	[:position,'%Position'],
	[:size,'%Size'],
	[:border_width,''],
	[:above,'Window'],
	[:override_redirect,'Bool'],
])

XGravityEvent = XAnyEvent.subtype([
	# first window was called event instead
	[:window,'Window'],
	[:position,'%Position'],
])

XResizeRequestEvent = XAnyEvent.subtype([
	[:size,'%Size'],
])

XConfigureRequestEvent = XAnyEvent.subtype([
	# first window was called parent instead
	[:window,'Window'],
	[:rectangle,'%Rectangle','position and size'],
	[:border_width,''],
	[:above,'Window'],
	[:detail,'','Above, Below, TopIf, BottomIf, Opposite'],
	[:value_mask,'ulong'],
])

XCirculateEvent = XAnyEvent.subtype([
	# first window was called event instead
	[:window,'Window'],
	[:place,'','PlaceOnTop, PlaceOnBottom'],
])

XCirculateRequestEvent = XAnyEvent.subtype([
	# first window was called parent instead
	[:window,'Window'],
	[:place,'','PlaceOnTop, PlaceOnBottom'],
])

XPropertyEvent = XAnyEvent.subtype([
	[:atom,'Atom'],
	[:time,'Time'],
	[:state,'','NewValue, Deleted'],
])

XSelectionClearEvent = XAnyEvent.subtype([
	[:selection,'Atom'],
	[:time,'Time'],
])

XSelectionRequestEvent = XAnyEvent.subtype([
	# first window was called owner instead
	[:requestor,'Window'],
	[:selection,'Atom'],
	[:target,'Atom'],
	[:property,'Atom'],
	[:time,'Time'],
])

XSelectionEvent = XAnyEvent.subtype([
	# first window was called requestor instead
	[:selection,'Atom'],
	[:target,'Atom'],
	[:property,'Atom','ATOM or None'],
	[:time,'Time'],
])

XColormapEvent = XAnyEvent.subtype([
	[:colormap,'Colormap','COLORMAP or None'],
	[:new,'Bool'],
	[:state,'','ColormapInstalled, ColormapUninstalled'],
])

XClientMessageEvent = XAnyEvent.subtype([
	[:message_type,'Atom'],
	[:format,''],
	[:data,'union { char b[20]; short s[10]; long l[5]; }'],
])

XMappingEvent = XAnyEvent.subtype([
	[:request,'','MappingModifier, MappingKeyboard, MappingPointer'],
	[:first_keycode,'','first keycode'],
	[:count,'','defines range of change w. first_keycod'],
])

XErrorEvent = C_Struct([
	[:type,''],
	DPY,
	[:resourceid,'XID','resource id'],
	[:serial,'ulong','serial number of failed request'],
	[:error_code,'uchar','error code of failed request'],
	[:request_code,'uchar','Major op-code of failed request'],
	[:minor_code,'uchar','Minor op-code of failed request'],
])

#define XAllocID(dpy) ((*((_XPrivDisplay)dpy)->resource_alloc)((dpy)))

XCharStruct = C_Struct(
	'per character font metric information.',
[
	[:lbearing,'short','origin to left edge of raster'],
	[:rbearing,'short','origin to right edge of raster'],
	[:width,'short',"advance to next char's origin"],
	[:ascent,'short','baseline to top edge of raster'],
	[:descent,'short','baseline to bottom edge of raster'],
	[:attributes,'ushort','per char flags (not predefined)'],
])

XFontProp = C_Struct(
	'To allow arbitrary information with fonts, there are additional properties
	returned.',
[
	[:name,'Atom'],
	[:card32,'ulong'],
])

XFontStruct = C_Struct([
	[:ext_data,'XExtData*','hook for extension to hang data'],
	[:fid,'Font','Font id for this font'],
	[:direction,'uint','hint about direction the font is painted'],
	[:min_char_or_byte2,'uint','first character'],
	[:max_char_or_byte2,'uint','last character'],
	[:min_byte1,'uint','first row that exists'],
	[:max_byte1,'uint','last row that exists'],
	[:all_chars_exist,'Bool','flag if all characters have non-zero siz'],
	[:default_char,'uint','char to print for undefined character'],
	[:n_properties,'','how many properties there are'],
	[:properties,'XFontProp*','pointer to array of additional propertie'],
	[:min_bounds,'XCharStruct','minimum bounds over all existing cha'],
	[:max_bounds,'XCharStruct','maximum bounds over all existing cha'],
	[:per_char,'XCharStruct*','first_char to last_char information'],
	[:ascent,'','log. extent above baseline for spacing'],
	[:descent,'','log. descent below baseline for spacing'],
])

XTextItem = C_Struct(
	'PolyText routines take these as arguments',
[
	[:chars,'char*','pointer to string'],
	[:nchars,'','number of characters'],
	[:delta,'','delta between strings'],
	[:font,'Font',"font to print it in, None don't change"],

])

XChar2b = C_Struct(
	'normal 16 bit characters are two bytes',
[
	[:byte1,'uchar'],
	[:byte2,'uchar'],
])

XTextItem16 = C_Struct([
	[:chars,'XChar2b*','two byte characters'],
	[:nchars,'','number of characters'],
	[:delta,'','delta between strings'],
	[:font,'Font',"font to print it in, None don't change"],
])

#!@#$
#typedef union {
#	DPY,
#	GC,
#	Visual *visual;
#	SCR,
#	ScreenFormat *pixmap_format;
#	XFontStruct *font;
#} XEDataObject;

XFontSetExtents = C_Struct([
	[:max_ink_extent,'XRectangle'],
	[:max_logical_extent,'XRectangle'],
])

#!@#$
#'unused:
#typedef void (*XOMProc)();
#'],

#!@#$
#typedef struct _XOM *XOM;
#typedef struct _XOC *XOC, *XFontSet;

XmbTextItem = C_Struct([
	[:chars,'char*'],
	[:nchars],
	[:delta],
	[:font_set,'XFontSet'],
])

XwcTextItem = C_Struct([
	[:chars,'wchar_t*'],
	[:nchars],
	[:delta],
	[:font_set,'XFontSet'],
])

XOMCharSetList = C_Struct([
	[:charset_count],
	[:charset_list,'char **'],
])

Enum(%w(
	XOMOrientation_LTR_TTB
	XOMOrientation_RTL_TTB
	XOMOrientation_TTB_LTR
	XOMOrientation_TTB_RTL
	XOMOrientation_Context
))

XOMOrientation = C_Struct([
	[:num_orientation],
	[:orientation,'XOrientation *','Input Text description'],
])

XOMFontInfo = C_Struct([
	[:num_font],
	[:font_struct_list,'XFontStruct **'],
	[:font_name_list,'char **'],
])

#!@#$
#typedef struct _XIM *XIM;
#typedef struct _XIC *XIC;

#!@#$
#typedef void (*XIMProc)(XIM,XPointer,XPointer);
#typedef Bool (*XICProc)(XIC,XPointer,XPointer);
#typedef void (*XIDProc)(DPY,XPointer,XPointer);

#!@#$
#typedef ulong XIMStyle;

XIMStyles = C_Struct([
	[:count_styles,'ushort'],
	[:supported_styles,'XIMStyle *'],
])

#!@#$
Foo1 = C_BitField(
# 1
 	:XIMPreeditArea,
	:XIMPreeditCallbacks,
	:XIMPreeditPosition,
	:XIMPreeditNothing,
# 10
	:XIMPreeditNone,
	nil,nil,nil,
# 100
	:XIMStatusArea,
	:XIMStatusCallbacks,
	:XIMStatusNothing,
	:XIMStatusNone,
nil)

#!@#$
Foo2 = Enum(1..4, %w(
	XLookupNone
	XLookupChars
	XLookupKeySym
	XLookupBoth
))

#!@#$
#typedef XPointer XVaNestedList;

XIMCallback = C_Struct([
	[:client_data,'XPointer'],
	[:callback,'XIMProc'],
])

XICCallback = C_Struct([
	[:client_data,'XPointer'],
	[:callback,'XICProc'],
])

#!@#$
#typedef ulong XIMFeedback;

#!@#$
Foo2 = C_BitSet(
# 0
	XIMReverse,
	XIMUnderline,
	XIMHighlight,
	nil,
# 4
	nil,
	XIMPrimary,
	XIMSecondary,
	XIMTertiary,
# 8
	XIMVisibleToForward,
	XIMVisibleToBackword,
	XIMVisibleToCenter,
nil)

XIMText = C_Struct([
	[:length,'ushort'],
	[:feedback,'XIMFeedback *'],
	[:encoding_is_wchar,'Bool'],
	[:string,'union {char *multi_byte;wchar_t *wide_char;}'],
])

#!@#$
#typedef ulong XIMPreeditState;

#!@#$
#def
#	XIMPreeditUnKnown	0L
#	XIMPreeditEnable	1L
#	XIMPreeditDisable	(1L<<1)
#end

XIMPreeditStateNotifyCallbackStruct = C_Struct([
	[:state,'XIMPreeditState'],
])

typedef ulong XIMResetState;

#!@#$
#def
#	XIMInitialState		1L
#	XIMPreserveState	(1L<<1)
#end

typedef ulong XIMStringConversionFeedback;

Foo3 = C_BitSet(
	:XIMStringConversionLeftEdge,
	:XIMStringConversionRightEdge,
	:XIMStringConversionTopEdge,
	:XIMStringConversionBottomEdge,
	:XIMStringConversionConcealed,
	:XIMStringConversionWrapped,
nil)

XIMStringConversionText = C_Struct([
	[:length,'ushort'],
	[:feedback,'XIMStringConversionFeedback*'],
	[:encoding_is_wchar,'Bool'],
	[:string,'union {char *mbs; wchar_t *wcs;}'],
])

#!@#$
#typedef ushort XIMStringConversionPosition;
#typedef ushort XIMStringConversionType;

Enum(1..4, %w(
	XIMStringConversionBuffer
	XIMStringConversionLine
	XIMStringConversionWord
	XIMStringConversionChar
))

#!@#$
#typedef ushort XIMStringConversionOperation;

Enum(1..2, %w(
	XIMStringConversionSubstitution
	XIMStringConversionRetrieval
))

#'This typo was present in pre-R6.4 versions.  It is defined here for
#compatibility purposes only.  It should not be used, and may not be
#present in future releases.
#'],
#define XIMStringConversionRetrival XIMStringConversionRetrieval

Enum(%w(
	XIMForwardChar  XIMBackwardChar
	XIMForwardWord  XIMBackwardWord
	XIMCaretUp      XIMCaretDown,
	XIMNextLine     XIMPreviousLine,
	XIMLineStart    XIMLineEnd, 
	XIMAbsolutePosition
	XIMDontChange
))

XIMStringConversionCallbackStruct = C_Struct([
	[:position,'XIMStringConversionPosition'],
	[:direction,'XIMCaretDirection'],
	[:operation,'XIMStringConversionOperation'],
	[:factor,'ushort'],
	[:text,'XIMStringConversionText*'],
])

XIMPreeditDrawCallbackStruct = C_Struct([
	[:caret,'','Cursor offset within pre-edit string'],
	[:chg_first,'','Starting change position'],
	[:chg_length,'','Length of the change in character count'],
	[:text,'XIMText*'],
])

XIMCaretStyle = Enum(
	[:XIMIsInvisible,'Disable caret feedback'], 
	[:XIMIsPrimary,'UI defined caret feedback'],
	[:XIMIsSecondary,'UI defined caret feedback'])

XIMPreeditCaretCallbackStruct = C_Struct([
	[:position,'','Caret offset within pre-edit string'],
	[:direction,'XIMCaretDirection','Caret moves direction'],
	[:style,'XIMCaretStyle','Feedback of the caret'],
])

XIMStatusDataType = Enum ([
	XIMTextType,
	XIMBitmapType
])
	
XIMStatusDrawCallbackStruct = C_Struct([
	[:type,'XIMStatusDataType'],
	[:data,'union {XIMText *text; Pixmap  bitmap;}'],
])

XIMHotKeyTrigger = C_Struct([
	[:keysym,'KeySym'],
	:modifier,
	:modifier_mask,
])

XIMHotKeyTriggers = C_Struct([
	:num_hot_key,
	[:key,'XIMHotKeyTrigger*'],
])

#!@#$
#typedef ulong  XIMHotKeyState;

Enum(1..2, %w(
	XIMHotKeyStateON
	XIMHotKeyStateOFF
))

XIMValuesList = C_Struct([
	[count_values,ushort],
	[supported_values,'char **'],
])

# _XFUNCPROTOBEGIN

XFontStruct *XLoadQueryFont(DPY,
	[:name,'const char *'])

XFontStruct *XQueryFont(DPY,
	[:font_ID,'XID'])

XTimeCoord *XGetMotionEvents(DPY,Window,
	[:start,'Time'],
	[:stop,'Time'],
	:nevents_return)

XModifierKeymap *XDeleteModifiermapEntry(
	[:modmap, 'XModifierKeymap*'],
	[:keycode_entry,'KeyCode'],
	:modifier)

XModifierKeymap	*XGetModifierMapping(DPY)

XModifierKeymap	*XInsertModifiermapEntry(
	[:modmap, 'XModifierKeymap*'],
	[:keycode_entry,'KeyCode'],
	:modifier)

XModifierKeymap *XNewModifiermap(
	:max_keys_per_mod)

XImage *XCreateImage(DPY,
	[:visual,'Visual*'],
	[:depth,'uint'],
	:format,
	:offset,
	[:data,'char*'],
	[:size,'%Size_uint'],
	:bitmap_pad,
	:bytes_per_line)

Status XInitImage('XImage*')

XImage *XGetImage(DPY,Drawable,
	[:position,'%Position'],
	[:size,'%Size_uint'],
	[:plane_mask,'ulong'],
	:format)

XImage *XGetSubImage(DPY,Drawable,
	[:position,'%Position'],
	[:size,'%Size_uint'],
	[:plane_mask,'ulong'],
	:format,
	[:dest_image,'XImage*'],
	[:dest,'%Position'])

# X function declarations.

Display *XOpenDisplay(
	[:display_name,'const char *'])

void XrmInitialize(void)

char *XFetchBytes(DPY,
	Return(:nbytes))

char *XFetchBuffer(DPY,
	Return(:nbytes),
	:buffer)

char *XGetAtomName(DPY,Atom)

Status XGetAtomNames(DPY,
	[:atoms,'Atom*'],
	:count,
	Return(:names,'char*'))

char *XGetDefault(DPY,
	[:program,'const char *'],
	[:option,'const char *'])

char *XDisplayName('const char *')

char *XKeysymToString(KeySym)

#!@#$
#int (*XSynchronize(DPY,Bool))(DPY)
#int (*XSetAfterFunction(DPY,int(*)(DPY)))(DPY)

# -------------------------------------------------------------------------- #

DPY = ['display','Display*','display']

Atom XInternAtom(DPY,
	[:atom_name,'const char *'],
	[:only_if_exists,'Bool'])

Status XInternAtoms(DPY,
	[:names,'char**'],
	:count,
	[:onlyIfExists,'Bool'],
	Return(Atom))

Colormap XCopyColormapAndFree(DPY,Colormap)
Colormap XCreateColormap(DPY,Window,
	'Visual*',
	:alloc)

Cursor XCreatePixmapCursor(DPY,
	[:source,'Pixmap'],
	[:mask,Pixmap],
	[:foreground_color,'XColor*'],
	[:background_color,'XColor*'],
	[:position,'%Position_uint']);

Cursor XCreateGlyphCursor(DPY,
	[:source_font,Font],
	[:mask_font,Font],
	[:source_char,'uint'],
	[:mask_char,'uint'],
	[:foreground_color,'XColor*'],
	[:background_color,'XColor*'])

Cursor XCreateFontCursor(DPY,
	[:shape,'uint'])

Font XLoadFont(DPY,
	[:name,'const char *'])

GC XCreateGC(DPY,Drawable,
	[:valuemask,'ulong'],
	[:values,'XGCValues*'])

GContext XGContextFromGC(GC)
void XFlushGC(DPY,GC)

Pixmap XCreatePixmap(DPY,Drawable,
	[:size,'%Size_uint'],
	[:depth,'uint'])

Pixmap XCreateBitmapFromData(DPY,Drawable,
	[:data,'const char *'],
	[:size,'%Size_uint'])

Pixmap XCreatePixmapFromBitmapData(DPY,Drawable,
	[:data,'char *'],
	[:size,'%Size_uint'],
	[:fg,'ulong'],
	[:bg,'ulong'],
	[:depth,'uint'])

Window XCreateSimpleWindow(DPY,
	[:parent,Window],
	[:position,'%Position'],
	[:size,'%Size_uint'],
	[:border_width,'uint'],
	[:border,'ulong'],
	[:background,'ulong'])

Window XGetSelectionOwner(DPY,
	[:selection, 'Atom'])

Window XCreateWindow(DPY,
	[:parent,Window],
	[:position,'%Position'],
	[:size,'%Size_uint'],
	[:border_width,'uint'],
	:depth,
	[:class,'uint'],
	[:visual, 'Visual*'],
	[:valuemask,'ulong'],
	[:attributes,'XSetWindowAttributes*'])

Colormap *XListInstalledColormaps(DPY,Window,
	Return(:num))

char **XListFonts(DPY,
	[:pattern,'const char *'],
	:maxnames,
	Return(:actual_count))

char **XListFontsWithInfo(DPY,
	[:pattern,'const char *'],
	:maxnames,
	Return(:count),
	Return([:info,'XFontStruct*']))

char **XGetFontPath(DPY,
	Return(:npaths))

char **XListExtensions(DPY,
	Return(:nextensions))

Atom *XListProperties(DPY,Window,
	Return(:num_prop))

XHostAddress *XListHosts(DPY,
	Return(:nhosts),
	Return([:state,'Bool']))

KeySym XKeycodeToKeysym(DPY,
	[:keycode,KeyCode],
	:index)

KeySym XLookupKeysym(
	[:key_event,'XKeyEvent*'],
	:index)

KeySym *XGetKeyboardMapping(DPY,
	[:first_keycode,KeyCode],
	:keycode_count,
	Return(:keysyms_per_keycode))

KeySym XStringToKeysym('const char *')

long XMaxRequestSize(DPY)
long XExtendedMaxRequestSize(DPY)
char *XResourceManagerString(DPY)
char *XScreenResourceString(SCR)
ulong XDisplayMotionBufferSize(DPY)
VisualID XVisualIDFromVisual('Visual*')

# multithread routines

Status XInitThreads(void)
void XLockDisplay(DPY)
void XUnlockDisplay(DPY)

# routines for dealing with extensions

XExtCodes *XInitExtension(DPY,
	[:name,'const char *'])

XExtCodes *XAddExtension(DPY);
XExtData *XFindOnExtensionList(
	[:structure,'XExtData**'],
	:number)

XExtData **XEHeadOfExtensionList(
	[:object,XEDataObject])

# these are routines for which there are also macros
Window XRootWindow(DPY,SCRN)
Window XDefaultRootWindow(DPY)
Window XRootWindowOfScreen(SCR)
Visual *XDefaultVisual(DPY,SCRN)
Visual *XDefaultVisualOfScreen(SCR)
GC XDefaultGC(DPY,SCRN)
GC XDefaultGCOfScreen(SCR)
ulong XBlackPixel(DPY,SCRN)
ulong XWhitePixel(DPY,SCRN)
ulong XAllPlanes(void)
ulong XBlackPixelOfScreen(SCR)
ulong XWhitePixelOfScreen(SCR)
ulong XNextRequest(DPY)
ulong XLastKnownRequestProcessed(DPY)
char *XServerVendor(DPY)
char *XDisplayString(DPY)
Colormap XDefaultColormap(DPY,SCRN)
Colormap XDefaultColormapOfScreen(SCR)
Display *XDisplayOfScreen(SCR)
SCR XScreenOfDisplay(DPY,SCRN)
SCR XDefaultScreenOfDisplay(DPY)
long XEventMaskOfScreen(SCR)
int XScreenNumberOfScreen(SCR)

# WARNING, this type not in Xlib spec
#!@#$
#typedef int (*XErrorHandler) (DPY,
#	XErrorEvent*	/* error_event */
#);

XErrorHandler XSetErrorHandler (XErrorHandler)

# WARNING, this type not in Xlib spec
#!@#$
# typedef int (*XIOErrorHandler) (DPY);

XIOErrorHandler XSetIOErrorHandler (XIOErrorHandler)

XPixmapFormatValues *XListPixmapFormats(DPY,
	Return(:count))

int *XListDepths(DPY,SCRN,
	Return(:count))

# ICCCM routines for things that don't require special include files;
# other declarations are given in Xutil.h
Status XReconfigureWMWindow(DPY,Window,
	SCRN,
	[:mask,'uint'],
	[:changes,'XWindowChanges*'])

Status XGetWMProtocols(DPY,Window,
	Return([:protocols,'Atom*']),
	Return(:count))

Status XSetWMProtocols(DPY,Window,
	[:protocols,'Atom*'],
	:count)

Status XIconifyWindow(DPY,Window,SCRN)
Status XWithdrawWindow(DPY,Window,SCRN)

Status XGetCommand(DPY,Window,
	Return(
		[:args,'[]char*']))

Status XGetWMColormapWindows(DPY,Window,
	Return([:windows,'Window*']),
	Return(:count))

Status XSetWMColormapWindows(DPY,Window,
	[:colormap_windows,'Window*'],
	:count)

void XFreeStringList(
	[:list,'char**'])

int XSetTransientForHint(DPY,Window,
	:prop_window,Window)

# ------------------------------------------------------------------------- #
# The following are given in alphabetical order

int XActivateScreenSaver(DPY)

int XAddHost(DPY,
	[:host,'XHostAddress*'])

int XAddHosts(DPY,
	[:hosta,'XHostAddress*'],
	:num_hosts)

int XAddToExtensionList(
	[:structure,'struct _XExtData**'],
	[:ext_data,'XExtData*'])

int XAddToSaveSet(DPY,Window)

Status XAllocColor(DPY,Colormap,
	[:screen_in_out,'XColor*'])

Status XAllocColorCells(DPY,Colormap,
	[:contig,Bool],
	[:plane_masks_return,'ulong*'],
	[:nplanes,'uint'],
	Return([:pixels,'ulong']),
	[:npixels,'uint'])

Status XAllocColorPlanes(DPY,Colormap,
	[:contig,Bool],
	Return([:pixels,'ulong']),
	:ncolors,
	:nreds,
	:ngreens,
	:nblues,
	Return(
		[:rmask,'ulong'],
		[:gmask,'ulong'],
		[:bmask,'ulong']))

Status XAllocNamedColor(DPY,Colormap,
	[:color_name,ZString],
	Return(
		[:screen_def,'XColor*'],
		[:exact_def,'XColor*']))

int XCellsOfScreen(SCR)

int XChangeActivePointerGrab(DPY,
	[:event_mask,'uint'],
	Cursor,
	Time)

int XChangeGC(DPY,GC,
	[:valuemask,'ulong'],
	[:values,'XGCValues*'])

int XChangeKeyboardControl(DPY,
	[:value_mask,'ulong'],
	[:values,'XKeyboardControl*'])

int XChangeKeyboardMapping(DPY,
	:first_keycode,
	:keysyms_per_keycode,
	[:keysyms,'KeySym*'],
	:num_codes)

int XChangePointerControl(DPY,
	[:do_accel,Bool],
	[:do_threshold,Bool],
	:accel_numerator,
	:accel_denominator,
	:threshold)

int XChangeProperty(DPY,Window,
	[:property,Atom],
	[:type,Atom],
	:format,
	:mode,
	[:data,'const uchar*'],
	:nelements)

int XChangeSaveSet(DPY,Window,:change_mode)

int XChangeWindowAttributes(DPY,Window,
	[:valuemask,'ulong'],
	[:attributes,'XSetWindowAttributes*'])

Bool XCheckIfEvent(DPY,
	Return(nil,'XEvent*'),
	[:predicate,'Bool (*) (DPY,XEvent*,XPointer)'],
	[:arg,XPointer])

Bool XCheckMaskEvent(DPY,
	[:event_mask,'long'],
	Return([:event,'XEvent*']))

Bool XCheckTypedEvent(DPY,
	:event_type,
	Return([:event,'XEvent*']))

Bool XCheckTypedWindowEvent(DPY,Window,
	:event_type,
	Return([:event,'XEvent*']))

Bool XCheckWindowEvent(DPY,Window,
	[:event_mask,'long'],
	Return([:event,'XEvent*']))

int XCirculateSubwindows(DPY,Window,:direction)
int XCirculateSubwindowsDown(DPY,Window)
int XCirculateSubwindowsUp(DPY,Window)

int XClearArea(DPY,Window,
	[:position,'%Position'],
	[:size,'%Size_uint'],
	[:exposures,Bool])

int XClearWindow(DPY,Window)

int XConfigureWindow(DPY,Window,
	[:value_mask,'uint'],
	[:values,'XWindowChanges*'])

int XConvertSelection(DPY,
	[:selection,Atom],
	[:target,Atom],
	[:property,Atom],
	[:requestor,Window],
	Time)

int XCopyArea(DPY,
	[:src,'Drawable'],
	[:dest,'Drawable'],
	GC,
	[:src,'%Position'],
	[:size,'%Size_uint'],
	[:dest,'%Position'])

int XCopyGC(DPY,
	[:src,GC],
	[:valuemask,'ulong'],
	[:dest,GC])

int XCopyPlane(DPY,
	[:src,'Drawable'],
	[:dest,'Drawable'],
	GC,
	[:src,'%Position'],
	[:size,'%Size_uint'],
	[:dest,'%Position'],
	[:plane,'ulong'])

int XDefaultDepth(DPY,SCRN)
int XDefaultDepthOfScreen(SCR)
int XDefineCursor(DPY,Window,Cursor)

int XDeleteProperty(DPY,Window,
	[:property,Atom])

int XDestroyWindow(DPY,Window)
int XDestroySubwindows(DPY,Window)
int XDoesBackingStore(SCR)
Bool XDoesSaveUnders(SCR)

int XDisplayCells(DPY,SCRN)
int XDisplayHeight(DPY,SCRN)
int XDisplayHeightMM(DPY,SCRN)

int XDisplayKeycodes(DPY,
	Return(
		:min_keycodes,
		:max_keycodes))

int XDisplayPlanes(DPY,SCRN)
int XDisplayWidth(DPY,SCRN)
int XDisplayWidthMM(DPY,SCRN)

int XDrawArc(DPY,Drawable,GC,
	[:position,'%Position'],
	[:size,'%Size_uint'],
	:angle1,
	:angle2)

int XDrawArcs(DPY,Drawable,GC,
	[:arcs,'[]Xarc*'])

int XDrawImageString(DPY,Drawable,GC,
	[:position,'%Position'],
	[:string,'%String'])

int XDrawImageString16(DPY,Drawable,GC,
	[:position,'%Position'],
	[:string,'%String16'])

int XDrawLine(DPY,Drawable,GC,:x1,:x2,:y1,:y2)

int XDrawLines(DPY,Drawable,GC,
	[:points,'[]%XPoint'],
	:mode)

int XDrawPoint(DPY,Drawable,GC,
	[:position,'%Position'])

int XDrawPoints(DPY,Drawable,GC,
	[:points,'[]%XPoint'],
	:mode)

int XDrawRectangle(DPY,Drawable,GC,
	[:position,'%Position'],
	[:size,'%Size_uint'])

int XDrawRectangles(DPY,Drawable,GC,
	[:rectangles,'[]%XRectangle'])

int XDrawSegments(DPY,Drawable,GC,
	[:segments,'[]%XSegment'])

int XDrawString(DPY,Drawable,GC,
	[:position,'%Position'],
	[:string,'%String'])

int XDrawString16(DPY,Drawable,GC,
	[:position,'%Position'],
	[:string,'%String16'])

int XDrawText(DPY,Drawable,GC,
	[:position,'%Position'],
	[:items,'[]XTextItem*'])

int XDrawText16(DPY,Drawable,GC,
	[:position,'%Position'],
	[:items,'[]XTextItem16'])

int XEventsQueued(DPY,:mode)

Status XFetchName(DPY,Window,
	Return([:window_name,'char*']))

int XFillArc(DPY,Drawable,GC,
	[:position,'%Position'],
	[:size,'%Size_uint'],
	:angle1,
	:angle2)

int XFillArcs(DPY,Drawable,GC,
	[:arcs,'[]%XArc'])

int XFillPolygon(DPY,Drawable,GC,
	[:points,'[]%XPoint'],
	:shape,
	:mode)

int XFillRectangle(DPY,Drawable,GC,
	[:position,'%Position'],
	[:size,'%Size_uint'])

int XFillRectangles(DPY,Drawable,GC,
	[:rectangles,'[]%XRectangle'])

int XForceScreenSaver(DPY,:mode)

int XFree([:data,'void*'])

int XFreeColormap(DPY,Colormap)

int XFreeColors(DPY,Colormap,
	[:pixels,'[]ulong'],
	[:planes,ulong])

int XFreeCursor(DPY,Cursor)

int XFreeExtensionList(
	[:list,'[]char*'])

int XFreeFont(DPY,'XFontStruct*')

int XFreeFontInfo(
	[:names,'char**'],
	[:free_info,'XFontStruct*'],
	:actual_count)

int XFreeFontNames(
	[:list,'[]char*'])

int XFreeFontPath(
	[:list,'[]char*'])

int XFreeGC(DPY,GC)

int XFreeModifiermap('XModifierKeymap*')

int XFreePixmap(DPY,Pixmap)

int XGeometry(DPY,SCRN,
	[:position,ZString],
	[:default_position,ZString],
	[:bwidth,'uint'],
	[:fsize,'%Size_uint'],
	[:adder,'%Position'],
	[:position_return,'%Position_return'],
	[:size_return,'%Size_return'])

#!@#$
=begin
int XGetErrorDatabaseText(DPY,
	[:name,ZString],
	[:message,ZString],
	[:default_string,ZString],
	char*		/* buffer_return */,
	:length)

int XGetErrorText(DPY,
	:code,
	char*		/* buffer_return */,
	:length)
=end

Bool XGetFontProperty(
	[:font_struct,'XFontStruct*'],
	[:atom,Atom],
	Return([:value,'ulong']))

Status XGetGCValues(DPY,GC,
	[:valuemask,'ulong'],
	Return([:values,'XGCValues*']))

Status XGetGeometry(DPY,Drawable,
	Return(
		[:root,'Window*'],
		[:position,'%Position'],
		[:size,'%Size_uint'],
		[:border_width,'uint'],
		[:depth,'uint']))

Status XGetIconName(DPY,Window,
	Return([:icon_name,'char*']))

int XGetInputFocus(DPY,
	Return(
		[:focus,Window],
		:revert_to))

int XGetKeyboardControl(DPY,
	Return([:values,'XKeyboardState*']))

int XGetPointerControl(DPY,
	Return(
		:accel_numerator,
		:accel_denominator,
		:threshold))

#!@#$
#int XGetPointerMapping(DPY,
#	uchar*	/* map_return */,
#	int			/* nmap */
#);

int XGetScreenSaver(DPY,
	Return(
		:timeout,
		:interval,
		:prefer_blanking_return,
		:allow_exposures))

Status XGetTransientForHint(DPY,Window,
	Return([:prop_window,Window]))

int XGetWindowProperty(DPY,Window,
	[:property,Atom],
	[:long_offset,'long'],
	[:long_offset,'long'],
	[:delete,Bool],
	[:req_type,Atom],
	Return(
		[:actual_type,Atom],
		:actual_format,
		[:nitems,'ulong'],
		[:bytes_after,'ulong'],
		[:prop,'char*']))

Status XGetWindowAttributes(DPY,Window,
	Return([:window_attributes,'XWindowAttributes*']))

PointerMode = int
KeyboardMode = int

int XGrabButton(DPY,
	[:button,'uint'],
	[:modifiers,'uint'],
	[:grab_window,Window],
	[:owner_events,Bool],
	[:event_mask,'uint'],
	PointerMode,
	KeyboardMode,
	[:confine_to,Window],
	Cursor)

int XGrabKey(DPY,
	:keycode,
	[:modifiers,'uint'],
	[:grab_window,Window],
	[:owner_events,Bool],
	PointerMode,
	KeyboardMode)

int XGrabKeyboard(DPY,
	[:grab_window,Window],
	[:owner_events,Bool],
	PointerMode,
	KeyboardMode,
	Time)

int XGrabPointer(DPY,
	[:grab_window,Window],
	[:owner_events,Bool],
	[:event_mask,'uint'],
	PointerMode,
	KeyboardMode,
	[:confine_to,Window],
	Cursor,
	Time)

int XHeightMMOfScreen(SCR)
int XHeightOfScreen(SCR)

int XIfEvent(DPY,
	Return([:event,'XEvent*']),
	[:predicate,'Bool (*) (DPY,XEvent*,XPointer)'],
	[:arg,XPointer])

int XInstallColormap(DPY,Colormap)

KeyCode XKeysymToKeycode(DPY,KeySym)
int XKillClient(DPY,
	[:resource,XID])

Status XLookupColor(DPY,Colormap,
	[:color_name,ZString],
	Return(
		[:screen_def,'XColor*'],
		[:exact_def,'XColor*']))

int XLowerWindow(DPY,Window)
int XMapRaised(DPY,Window)
int XMapSubwindows(DPY,Window)
int XMapWindow(DPY,Window)

int XMaskEvent(DPY,
	[:event_mask,'long'],
	Return([:event,'XEvent*']))

int XMaxCmapsOfScreen(SCR)
int XMinCmapsOfScreen(SCR)

int XMoveResizeWindow(DPY,Window,
	[:position,'%Position'],
	[:size,'%Size_uint'])

int XMoveWindow(DPY,Window,
	[:position,'%Position'])

int XNextEvent(DPY,
	Return([:event,'XEvent*']))

Status XParseColor(DPY,
	Colormap,
	[:spec,ZString],
	Return(
		[:exact_def,'XColor*']))

int XParseGeometry(
	[:parsestring,ZString],
	[:position_return,'%Position_return'],
	[:size_return,'%Size_return_uint'])

int XPeekEvent(DPY,
	Return([:event,'XEvent*']))

int XPeekIfEvent(DPY,
	Return([:event,'XEvent*']),
	[:predicate,'Bool (*) (DPY,XEvent*,XPointer)'],
	[:arg,XPointer])

int XPlanesOfScreen(SCR)

int XPutBackEvent(DPY,'XEvent*')

int XPutImage(DPY,Drawable,GC,'XImage*',
	[:src,'%Position'],
	[:dest,'%Position'],
	[:size,'%Size_uint']);

Status XQueryBestCursor(DPY,Drawable,
	[:size,'%Size_uint'],
	[:size_return,'%Size_return_uint'])

Status XQueryBestSize(DPY,
	:class,
	[:which_screen,Drawable],
	[:size,'%Size_uint'],
	[:size_return,'%Size_return_uint'])

Status XQueryBestStipple(DPY,
	[:which_screen,Drawable],
	[:size,'%Size_uint'],
	[:size_return,'%Size_return_uint'])

Status XQueryBestTile(DPY,
	[:which_screen,Drawable],
	[:size,'%Size_uint'],
	[:size_return,'%Size_return_uint'])

int XQueryColor(DPY,Colormap,
	[:def_in_out,'XColor*'])

int XQueryColors(DPY,Colormap,
	[:defs_in_out,'[]XColor*'])

Bool XQueryExtension(DPY,
	[:name,ZString],
	Return(
		:major_opcode,
		:first_event,
		:first_error))

int XQueryKeymap(DPY,
	Return(:keys,'char[32]'))

Bool XQueryPointer(DPY,Window,
	Return(
		[:root,'Window*'],
		[:child,'Window*'],
		[:root_position,'%Position'],
		[:win_position,'%Position'],
		[:mask,'uint']))

int XQueryTextExtents(DPY,
	[:font_ID,XID],
	[:string,'%String'],
	Return(
		[:direction_return],
		[:font_ascent_return],
		[:font_descent_return],
		[:overall,'XCharStruct*']))

int XQueryTextExtents16(DPY,
	[:font_ID,XID],
	[:string,'%String16'],
	Return(
		[:direction],
		[:font_ascent],
		[:font_descent],
		[:overall,'XCharStruct*']))

Status XQueryTree(DPY,Window,
	Return(
		[:root,Window],
		[:parent,Window],
		[:children,Window]))

int XRaiseWindow(DPY,Window)

int XReadBitmapFile(DPY,Drawable,
	[:filename,ZString],
	Return(
		[:size,'%Size_uint'],
		[:bitmap,'Pixmap*'],
		[:hot,'%Position']))

int XReadBitmapFileData(
	[:filename,ZString],
	Return(
		[:size,'%Size_uint'],
		[:data,'uchar*'],
		[:hot,'%Position']))

int XRebindKeysym(DPY,
	[:keysym,KeySym],
	[:list,'[]KeySym'],
	:mod_count,
	[:string,'%String(uchar)'])

int XRecolorCursor(DPY,
	Cursor,
	[:foreground_color,'XColor*'],
	[:background_color,'XColor*'])

int XRefreshKeyboardMapping(
	[:event_map,'XMappingEvent*'])

int XRemoveFromSaveSet(DPY,Window)

int XRemoveHost(DPY,
	[:host,'XHostAddress*'])

int XRemoveHosts(DPY,
	[:hosts,'XHostAddress*'],
	:num_hosts)

int XReparentWindow(DPY,Window,
	[:parent,'Window'],
	[:position,'%Position'])

int XResizeWindow(DPY,Window,
	[:size,'%Size_uint'])

int XRestackWindows(DPY,
	[:windows,'[]Window'])

int XRotateBuffers(DPY,:rotate)

int XRotateWindowProperties(DPY,Window,
	[:properties,'[]Atom'],
	:npositions)

int XSelectInput(DPY,Window,
	[:event_mask,'long'])

Status XSendEvent(DPY,Window,
	[:propagate,Bool],
	[:event_mask,'long'],
	[:event_send,'XEvent*'])

int XSetAccessControl(DPY,:mode)

int XSetArcMode(DPY,GC,:arc_mode)

int XSetBackground(DPY,GC,
	[:background,'ulong'])

int XSetClipMask(DPY,GC,Pixmap)

int XSetClipOrigin(DPY,GC,
	[:clip_origin,'%Position'])

int XSetClipRectangles(DPY,GC,
	[:clip_origin,'%Position'],
	[:rectangles,'[]XRectangle'],
	:ordering)

int XSetCloseDownMode(DPY,:close_mode)

int XSetCommand(DPY,Window,
	[:args,'[]char*'])

int XSetDashes(DPY,GC,
	:dash_offset,
	[:dash_list,'[]const char'])

int XSetFillRule(DPY,GC,:fill_rule)
int XSetFillStyle(DPY,GC,:fill_style)

int XSetFont(DPY,GC,Font)

int XSetFontPath(DPY,
	[:directories,'[]char*'])

int XSetForeground(DPY,GC,
	[:foreground,'ulong'])

int XSetFunction(DPY,GC,
	:function)

int XSetGraphicsExposures(DPY,GC,
	[:graphics_exposures,Bool])

int XSetIconName(DPY,Window,
	[:icon_name,ZString])

int XSetInputFocus(DPY,
	[:focus,Window],
	:revert_to,
	Time)

int XSetLineAttributes(DPY,GC,
	[:line_width,'uint'],
	:line_style,
	:cap_style,
	:join_style)

int XSetModifierMapping(DPY,
	[:modmap,'XModifierKeymap*'])

int XSetPlaneMask(DPY,GC,
	[:plane_mask,'ulong'])

int XSetPointerMapping(DPY,
	[:map,'[]const uchar'])

int XSetScreenSaver(DPY,
	:timeout,
	:interval,
	:prefer_blanking,
	:allow_exposures)

int XSetSelectionOwner(DPY,
	[:selection,Atom],
	[:owner,Window],
	Time)

int XSetState(DPY,GC,
	[:foreground,'ulong'],
	[:background,'ulong'],
	:function,
	[:plane_mask,'ulong'])

int XSetStipple(DPY,GC,
	[:stipple,Pixmap])

int XSetSubwindowMode(DPY,GC,
	:subwindow_mode)

int XSetTSOrigin(DPY,GC,
	[:ts_origin,'%Position'])

int XSetTile(DPY,GC,
	[:tile,Pixmap])

int XSetWindowBackground(DPY,Window,
	[:background_pixel,'ulong'])

int XSetWindowBackgroundPixmap(DPY,Window,
	[:background_pixmap,Pixmap])

int XSetWindowBorder(DPY,Window,
	[:border_pixel,'ulong'])

int XSetWindowBorderPixmap(DPY,Window,
	[:border_pixmap,Pixmap])

int XSetWindowBorderWidth(DPY,Window,
	[:width,'uint'])

int XSetWindowColormap(DPY,Window,Colormap)

int XStoreBuffer(DPY,
	[:bytes,'%Bytes'],
	:buffer)

int XStoreBytes(DPY,
	[:bytes,'%Bytes'])

int XStoreColor(DPY,Colormap,
	[:color,'XColor*'])

int XStoreColors(DPY,Colormap,
	[:color,'[]XColor*'])

int XStoreName(DPY,Window,
	[:window_name,ZString])

int XStoreNamedColor(DPY,Colormap,
	[:color,ZString],
	[:pixel,'ulong'],
	:flags)

int XSync(DPY,
	[:discard,Bool])

int XTextExtents('XFontStruct*',
	[:string,'%String'],
	Return(
		:direction,
		:font_ascent,
		:font_descent,
		[:overall,'XCharStruct*']))

int XTextExtents16('XFontStruct*',
	[:string,'%String16'],
	Return(
		:direction,
		:font_ascent,
		:font_descent,
		[:overall,'XCharStruct*']))

int XTextWidth('XFontStruct*',
	[:string,'%String'])

int XTextWidth16('XFontStruct*',
	[:string,'%String16'])

Bool XTranslateCoordinates(DPY,
	[:src_w,Window],
	[:dest_w,Window],
	[:src,'%Position'],
	Return(
		[:dest,'%Position'],
		[:child,Window]))

int XUndefineCursor(DPY,Window)

int XUngrabButton(DPY,
	[:button,'uint'],
	[:modifiers,'uint'],
	[:grab_window,Window])

int XUngrabKey(DPY,
	:keycode,
	[:modifiers,'uint'],
	[:grab_window,Window])

int XUngrabKeyboard(DPY,Time)
int XUngrabPointer(DPY,Time)

int XUninstallColormap(DPY,Colormap)

int XUnloadFont(DPY,Font)

int XUnmapSubwindows(DPY,Window)
int XUnmapWindow(DPY,Window)

int XWarpPointer(DPY,
	[:src_w,Window],
	[:dest_w,Window],
	[:src,'%Position'],
	[:size,'%Size_uint'],
	[:dest,'%Position'])

int XWidthMMOfScreen(SCR)
int XWidthOfScreen(SCR)

int XWindowEvent(DPY,Window,
	[:event_mask,'long'],
	Return(nil,'XEvent*'))

int XWriteBitmapFile(DPY,
	[:filename,ZString],
	[:bitmap,Pixmap],
	[:size,'%Size_uint'],
	[:hot,'%Position'])

Bool XSupportsLocale(void)

char *XSetLocaleModifiers(
	[:modifier_list,ZString])

XOM XOpenOM(DPY,
	[:rdb,'struct _XrmHashBucketRec*'],
	[:res_name,ZString],
	[:res_class,ZString])

#!@#$
#Status XCloseOM(XOM)
#char *XSetOMValues(XOM, ...)
#char *XGetOMValues(XOM, ...)
#Display *XDisplayOfOM(XOM)
#char *XLocaleOfOM(XOM)
#XOC XCreateOC(XOM, ...)
#void XDestroyOC(XOC)
#XOM XOMOfOC(XOC)
#char *XSetOCValues(XOC)
#char *XGetOCValues(XOC, ...)

XFontSet XCreateFontSet(DPY,
	[:base_font_name,ZString],
	[:missing_charset_list,'[]char**'],
	[:def_string,'char**'])

void XFreeFontSet(DPY,XFontSet)

int XFontsOfFontSet(XFontSet,
	[:font_struct_list,'XFontStruct***'],
	[:font_name_list,'char***'])

char *XBaseFontNameListOfFontSet(XFontSet)
char *XLocaleOfFontSet(XFontSet)
Bool XContextDependentDrawing(XFontSet)
Bool XDirectionalDependentDrawing(XFontSet)
Bool XContextualDrawing(XFontSet)
XFontSetExtents *XExtentsOfFontSet(XFontSet)

int XmbTextEscapement(XFontSet,
	[:text,'%String'])

int XwcTextEscapement(XFontSet,
	[:text,WString])

int XmbTextExtents(XFontSet,
	[:text,ZString],
	Return(
		[:overall_ink,'XRectangle*'],
		[:overall_logical,'XRectangle*']))

int XwcTextExtents(XFontSet,
	[:text,WString],
	Return(
		[:overall_ink,'XRectangle*'],
		[:overall_logical,'XRectangle*']))

Status XmbTextPerCharExtents(XFontSet,
	[:text,'%String'],
	[:ink_extents_buffer,'XRectangle*'],
	[:logical_extents_buffer,'XRectangle*'],
	:buffer_size,
	[:num_chars,'int*'],
	Return(
		[:overall_ink,'XRectangle*'],
		[:overall_logical,'XRectangle*']))

Status XwcTextPerCharExtents(XFontSet,
	[:text,WString],
	[:ink_extents_buffer,'XRectangle*'],
	[:logical_extents_buffer,'XRectangle*'],
	:buffer_size,
	[:num_chars,'int*'],
	Return(
		[:overall_ink,'XRectangle*'],
		[:overall_logical,'XRectangle*']))

void XmbDrawText(DPY,Drawable,GC,
	[:position,'%Position'],
	[:text_items,'[]XmbTextItem'])

void XwcDrawText(DPY,Drawable,GC,
	[:position,'%Position'],
	[:text_items,'[]XwcTextItem*'])

void XmbDrawString(DPY,Drawable,XFontSet,GC,
	[:position,'%Position'],
	[:text,'%String'])

void XwcDrawString(DPY,Drawable,XFontSet,GC,
	[:position,'%Position'],
	[:text,WString])

void XmbDrawImageString(DPY,Drawable,XFontSet,GC,
	[:position,'%Position'],
	[:text,'%String'])

void XwcDrawImageString(DPY,Drawable,XFontSet,GC,
	[:position,'%Position'],
	[:text,WString])

XIM XOpenIM(DPY,
	[:rdb,'struct _XrmHashBucketRec*'],
	[:res_name,'char*'],
	[:res_class,'char*'])

Status XCloseIM(XIM)

#!@#$
#char *XGetIMValues(XIM, ...)
#char *XSetIMValues(XIM, ...)
#Display *XDisplayOfIM(XIM)
#char *XLocaleOfIM(XIM)
#XIC XCreateIC(XIM, ...)

void XDestroyIC(XIC)
void XSetICFocus(XIC)
void XUnsetICFocus(XIC)
wchar_t *XwcResetIC(XIC)
char *XmbResetIC(XIC)

#!@#$
#char *XSetICValues(XIC, ...)
#char *XGetICValues(XIC, ...)
#XIM XIMOfIC(XIC)

Bool XFilterEvent('XEvent*',Window)

int XmbLookupString(XIC,
	[:event,'XKeyPressedEvent*'],
	Return(
		[:buffer,'%String'],
		[:keysym,'KeySym*'],
		[:status,'Status*']))

int XwcLookupString(XIC,
	[:event,'XKeyPressedEvent*'],
	Return(
		[:buffer,WString],
		[:keysym,'KeySym*'],
		[:status,'Status*']))

#!@#$
#XVaNestedList XVaCreateNestedList(:unused, ...)

# internal connections for IMs

Bool XRegisterIMInstantiateCallback(DPY,'_XrmHashBucketRec*',
	[:res_name,'char*'],
	[:res_class,'char*'],
	[:callback,XIDProc],
	[:client_data,XPointer])

Bool XUnregisterIMInstantiateCallback(DPY,'_XrmHashBucketRec*',
	[:res_name,'char*'],
	[:res_class,'char*'],
	[:callback,XIDProc],
	[:client_data,XPointer])

#!@#$
#typedef void (*XConnectionWatchProc)(DPY,
#	XPointer			/* client_data */,
#	:fd,
#	Bool			/* opening */,	 /* open or close flag */
#	XPointer*			/* watch_data */ /* open sets, close uses */
#)

Status XInternalConnectionNumbers(DPY,
	Return(
		[:fd,'[]int']))

void XProcessInternalConnection(DPY,:fd)

Status XAddConnectionWatch(DPY,
	[:callback,XConnectionWatchProc],
	[:client_data,XPointer])

void XRemoveConnectionWatch(DPY,
	[:callback,XConnectionWatchProc],
	[:client_data,XPointer])

_XFUNCPROTOEND

=begin
#def
 Bool int
 Status int
 True 1
 False 0

 QueuedAlready 0
 QueuedAfterReading 1
 QueuedAfterFlush 2

 ConnectionNumber(dpy) 	(((_XPrivDisplay)dpy)->fd)
 RootWindow(dpy, scr) 	(ScreenOfDisplay(dpy,scr)->root)
 DefaultScreen(dpy) 	(((_XPrivDisplay)dpy)->default_screen)
 DefaultRootWindow(dpy) 	(ScreenOfDisplay(dpy,DefaultScreen(dpy))->root)
 DefaultVisual(dpy, scr) (ScreenOfDisplay(dpy,scr)->root_visual)
 DefaultGC(dpy, scr) 	(ScreenOfDisplay(dpy,scr)->default_gc)
 BlackPixel(dpy, scr) 	(ScreenOfDisplay(dpy,scr)->black_pixel)
 WhitePixel(dpy, scr) 	(ScreenOfDisplay(dpy,scr)->white_pixel)
 AllPlanes 		((ulong)~0L)
 QLength(dpy) 		(((_XPrivDisplay)dpy)->qlen)
 DisplayWidth(dpy, scr) 	(ScreenOfDisplay(dpy,scr)->width)
 DisplayHeight(dpy, scr) (ScreenOfDisplay(dpy,scr)->height)
 DisplayWidthMM(dpy, scr)(ScreenOfDisplay(dpy,scr)->mwidth)
 DisplayHeightMM(dpy, scr)(ScreenOfDisplay(dpy,scr)->mheight)
 DisplayPlanes(dpy, scr) (ScreenOfDisplay(dpy,scr)->root_depth)
 DisplayCells(dpy, scr) 	(DefaultVisual(dpy,scr)->map_entries)
 ScreenCount(dpy) 	(((_XPrivDisplay)dpy)->nscreens)
 ServerVendor(dpy) 	(((_XPrivDisplay)dpy)->vendor)
 ProtocolVersion(dpy) 	(((_XPrivDisplay)dpy)->proto_major_version)
 ProtocolRevision(dpy) 	(((_XPrivDisplay)dpy)->proto_minor_version)
 VendorRelease(dpy) 	(((_XPrivDisplay)dpy)->release)
 DisplayString(dpy) 	(((_XPrivDisplay)dpy)->display_name)
 DefaultDepth(dpy, scr) 	(ScreenOfDisplay(dpy,scr)->root_depth)
 DefaultColormap(dpy, scr)(ScreenOfDisplay(dpy,scr)->cmap)
 BitmapUnit(dpy) 	(((_XPrivDisplay)dpy)->bitmap_unit)
 BitmapBitOrder(dpy) 	(((_XPrivDisplay)dpy)->bitmap_bit_order)
 BitmapPad(dpy) 		(((_XPrivDisplay)dpy)->bitmap_pad)
 ImageByteOrder(dpy) 	(((_XPrivDisplay)dpy)->byte_order)
 NextRequest(dpy)	(((_XPrivDisplay)dpy)->request + 1)
 LastKnownRequestProcessed(dpy)	(((_XPrivDisplay)dpy)->last_request_read)
#end

/* macros for screen oriented applications (toolkit) */
#def
 ScreenOfDisplay(dpy, scr)(&((_XPrivDisplay)dpy)->screens[scr])
 DefaultScreenOfDisplay(dpy) ScreenOfDisplay(dpy,DefaultScreen(dpy))
 DisplayOfScreen(s)	((s)->display)
 RootWindowOfScreen(s)	((s)->root)
 BlackPixelOfScreen(s)	((s)->black_pixel)
 WhitePixelOfScreen(s)	((s)->white_pixel)
 DefaultColormapOfScreen(s)((s)->cmap)
 DefaultDepthOfScreen(s)	((s)->root_depth)
 DefaultGCOfScreen(s)	((s)->default_gc)
 DefaultVisualOfScreen(s)((s)->root_visual)
 WidthOfScreen(s)	((s)->width)
 HeightOfScreen(s)	((s)->height)
 WidthMMOfScreen(s)	((s)->mwidth)
 HeightMMOfScreen(s)	((s)->mheight)
 PlanesOfScreen(s)	((s)->root_depth)
 CellsOfScreen(s)	(DefaultVisualOfScreen((s))->map_entries)
 MinCmapsOfScreen(s)	((s)->min_maps)
 MaxCmapsOfScreen(s)	((s)->max_maps)
 DoesSaveUnders(s)	((s)->save_unders)
 DoesBackingStore(s)	((s)->backing_store)
 EventMaskOfScreen(s)	((s)->root_input_mask)
#end

=end

#def
 XNRequiredCharSet "requiredCharSet"
 XNQueryOrientation "queryOrientation"
 XNBaseFontName "baseFontName"
 XNOMAutomatic "omAutomatic"
 XNMissingCharSet "missingCharSet"
 XNDefaultString "defaultString"
 XNOrientation "orientation"
 XNDirectionalDependentDrawing "directionalDependentDrawing"
 XNContextualDrawing "contextualDrawing"
 XNFontInfo "fontInfo"
#end

 XNVaNestedList "XNVaNestedList"
 XNQueryInputStyle "queryInputStyle"
 XNClientWindow "clientWindow"
 XNInputStyle "inputStyle"
 XNFocusWindow "focusWindow"
 XNResourceName "resourceName"
 XNResourceClass "resourceClass"
 XNGeometryCallback "geometryCallback"
 XNDestroyCallback "destroyCallback"
 XNFilterEvents "filterEvents"
 XNPreeditStartCallback "preeditStartCallback"
 XNPreeditDoneCallback "preeditDoneCallback"
 XNPreeditDrawCallback "preeditDrawCallback"
 XNPreeditCaretCallback "preeditCaretCallback"
 XNPreeditStateNotifyCallback "preeditStateNotifyCallback"
 XNPreeditAttributes "preeditAttributes"
 XNStatusStartCallback "statusStartCallback"
 XNStatusDoneCallback "statusDoneCallback"
 XNStatusDrawCallback "statusDrawCallback"
 XNStatusAttributes "statusAttributes"
 XNArea "area"
 XNAreaNeeded "areaNeeded"
 XNSpotLocation "spotLocation"
 XNColormap "colorMap"
 XNStdColormap "stdColorMap"
 XNForeground "foreground"
 XNBackground "background"
 XNBackgroundPixmap "backgroundPixmap"
 XNFontSet "fontSet"
 XNLineSpace "lineSpace"
 XNCursor "cursor"

 XNQueryIMValuesList "queryIMValuesList"
 XNQueryICValuesList "queryICValuesList"
 XNVisiblePosition "visiblePosition"
 XNR6PreeditCallback "r6PreeditCallback"
 XNStringConversionCallback "stringConversionCallback"
 XNStringConversion "stringConversion"
 XNResetState "resetState"
 XNHotKey "hotKey"
 XNHotKeyState "hotKeyState"
 XNPreeditState "preeditState"
 XNSeparatorofNestedList "separatorofNestedList"

 XBufferOverflow		-1

#end

